import abc


class DIException(Exception):
    def __init__(self, contract):
        self.contract = contract
        super().__init__(self.message())

    @abc.abstractmethod
    def message(self):
        pass


class ContractIsNotRegistered(DIException):
    def message(self):
        return f'Contract is not registered: {self.contract}'


class CannotAutomaticallyDeriveContract(DIException):
    def message(self):
        return f'Cannot automatically derive contract: {self.contract}'



import collections.abc
import types
import typing

from .. import contracts


class Resolver:
    def __init__(
        self,
        unit_registry: contracts.internal.ContainerUnitRegistry,
        callback_registry: collections.abc.MutableSequence[contracts.internal.ContainerUnit],
        global_cache: contracts.internal.InstanceCacheType | None = None
    ):
        self._unit_registry = unit_registry
        self._callback_registry = callback_registry
        self._global_cache = global_cache or {}
        self._local_cache: contracts.internal.InstanceCacheType = {}
        self._context = None

    def add_context(
        self,
        contract: contracts.ContractType,
        data: collections.abc.Mapping
    ):
        self._context = self._context or {contract: data}

    async def resolve(
        self,
        contract: contracts.ContractType,
        execution_context: collections.abc.Mapping[str, typing.Any] | None = None,
        initial_context: collections.abc.Mapping[str, typing.Any] | None = None
    ):
        return await self._resolve(
            contract=contract,
            execution_context=execution_context,
            initial_context=initial_context
        )

    async def shutdown(self, exc_type, exc_val):
        for callback_unit in self._callback_registry:
            if callback_unit.cache_scope is contracts.CacheScope.RESOLVER:
                instance = self._local_cache.get(callback_unit.contract)
                if callback_unit.implementation_execution_type is contracts.internal.ImplementationExecutionType.ASYNC:
                    await callback_unit.implementation(instance, {'exc_type': exc_type, 'exc_val': exc_val})
                else:
                    callback_unit.implementation(instance, {'exc_type': exc_type, 'exc_val': exc_val})

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.shutdown(exc_type, exc_val)

    async def _resolve(
        self,
        contract: contracts.ContractType,
        default: typing.Any | None = None,
        execution_context: collections.abc.Mapping[str, typing.Any] | None = None,
        initial_context: collections.abc.Mapping[str, typing.Any] | None = None
    ):
        if self._context and contract in self._context:
            return self._context.get(contract)
        if contract in self._local_cache:
            return self._local_cache.get(contract)
        if contract in self._global_cache:
            return self._global_cache.get(contract)
        if (unit := self._unit_registry.get(contract)) is None:
            return default

        result = None
        if unit.implementation_kind is contracts.internal.ImplementationKind.STATIC:
            result = unit.implementation
        elif unit.implementation_kind is contracts.internal.ImplementationKind.CLASS:
            params = {
                k: await self._apply_context_value(k, v, initial_context)
                for k, v in unit.init_signature.parameters.items()
                if k not in ('args', 'kwargs')
            } if unit.init_signature else {}
            result = unit.implementation(**params)
        elif unit.implementation_kind is contracts.internal.ImplementationKind.CALLABLE:
            params = {
                k: await self._apply_context_value(k, v, execution_context)
                for k, v in unit.execution_signature.parameters.items()
            } if unit.execution_signature else {}
            if unit.implementation_execution_type is contracts.internal.ImplementationExecutionType.ASYNC:
                result = await unit.implementation(**params)
            else:
                result = unit.implementation(**params)

        if result is not None:
            if unit.cache_scope is contracts.CacheScope.RESOLVER:
                self._local_cache[contract] = result
            elif unit.cache_scope is contracts.CacheScope.CONTAINER:
                self._global_cache[contract] = result

        return result

    async def _apply_context_value(
        self,
        key: str,
        value: contracts.internal.FunctionParameter,
        context: collections.abc.Mapping[str, typing.Any]
    ):
        if context and key in context:
            return context.get(key)

        args = list(filter(lambda x: x is not types.NoneType, typing.get_args(value.type)))
        if len(args) > 1:
            raise Exception("Cannot resolve more than 1 contract")
        elif len(args) == 1:
            _type = args[0]
        else:
            _type = value.type

        return await self._resolve(_type, value.default)



import collections.abc
import functools
import inspect
import typing
import itertools

from .. import contracts, exceptions
from . import resolver


class Container:
    def __init__(self):
        self._unit_registry: contracts.internal.ContainerUnitRegistry = {}
        self._callback_registry: collections.abc.MutableSequence[contracts.internal.ContainerUnit] = []
        self._cache: contracts.internal.InstanceCacheType = {}

    def register_dependency(
        self,
        implementation: contracts.internal.ImplementationType,
        contract: contracts.internal.ContractType | None = None,
        cache_scope: typing.Literal['container', 'resolver'] | None = None,
    ):
        if cache_scope == 'container':
            cache_scope = contracts.CacheScope.CONTAINER
        elif cache_scope == 'resolver':
            cache_scope = contracts.CacheScope.RESOLVER
        contract = self._derive_implementation_contract(implementation, contract)
        implementation_kind = self._derive_implementation_kind(implementation)
        implementation_execution_type = self._derive_implementation_execution_type(implementation)
        init_signature = None
        execution_signature = None
        if implementation_kind is contracts.internal.ImplementationKind.CLASS:
            init_signature = self._derive_parameters(getattr(implementation, '__init__'))
        elif implementation_kind is contracts.internal.ImplementationKind.CALLABLE:
            execution_signature = self._derive_parameters(
                implementation if inspect.isfunction(implementation) else getattr(implementation, '__call__')
            )

        self._unit_registry[contract] = contracts.internal.ContainerUnit(
            contract=contract,
            implementation=implementation,
            implementation_kind=implementation_kind,
            implementation_execution_type=implementation_execution_type,
            cache_scope=cache_scope,
            init_signature=init_signature,
            execution_signature=execution_signature
        )

    def register_callback(
        self,
        contract: contracts.internal.ContractType,
        callback: contracts.CallbackType,
        trigger: typing.Literal['shutdown'],
    ):
        if trigger == 'shutdown':
            trigger = contracts.ImplementationExecutionTrigger.SHUTDOWN

        execution_signature = self._derive_parameters(
            callback if inspect.isfunction(callback) else getattr(callback, '__call__')
        )
        self._callback_registry.append(
            contracts.internal.ContainerUnit(
                contract=contract,
                implementation=callback,
                implementation_kind=contracts.internal.ImplementationKind.CALLBACK,
                implementation_execution_type=self._derive_implementation_execution_type(callback),
                implementation_execution_trigger=trigger,
                cache_scope=self._unit_registry.get(contract).cache_scope,
                execution_signature=execution_signature
            )
        )

    def get_resolver(self) -> resolver.Resolver:
        return resolver.Resolver(
            unit_registry=self._unit_registry,
            callback_registry=self._callback_registry,
            global_cache=self._cache
        )

    async def shutdown(self, exc_type: typing.Any | None = None):
        pass

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.shutdown(exc_type)

    def __iter__(self) -> collections.abc.Generator[contracts.internal.ContainerUnit, None, None]:
        for unit in itertools.chain(self._unit_registry.values(), self._callback_registry):
            yield unit

    def __add__(self, other: typing.Self) -> typing.Self:
        for unit in other:
            if unit.implementation_kind == contracts.internal.ImplementationKind.CALLBACK:
                self._callback_registry.append(unit)
            else:
                self._unit_registry[unit.contract] = unit
        return self

    @staticmethod
    def _derive_implementation_contract(
        implementation: contracts.ImplementationType,
        contract: contracts.ContractType | None = None
    ):
        if contract is None:
            if inspect.isfunction(implementation):
                signature = inspect.signature(implementation)
                return signature.return_annotation
            elif inspect.isclass(implementation):
                try:
                    return implementation.__mro__[1]
                except IndexError:
                    raise exceptions.CannotAutomaticallyDeriveContract(implementation)
            else:
                raise exceptions.CannotAutomaticallyDeriveContract(implementation)
        return contract

    @staticmethod
    def _derive_parameters(
        func: typing.Callable
    ):
        if not inspect.isfunction(func):
            return None
        signature = inspect.signature(func)
        return contracts.internal.FunctionSignature(
            parameters={
                param_name: contracts.internal.FunctionParameter(
                    type=param_type.annotation,
                    default=None if param_type.default is inspect.Parameter.empty else param_type.default
                )
                for param_name, param_type
                in signature.parameters.items()
                if param_name != 'self'
            },
            result=signature.return_annotation
        )

    @staticmethod
    def _derive_implementation_kind(
        implementation: contracts.ImplementationType
    ) -> contracts.internal.ImplementationKind:
        if inspect.isclass(implementation):
            return contracts.internal.ImplementationKind.CLASS
        elif inspect.iscoroutinefunction(implementation):
            return contracts.internal.ImplementationKind.CALLABLE
        elif inspect.isfunction(implementation):
            return contracts.internal.ImplementationKind.CALLABLE
        elif isinstance(implementation, functools.partial):
            return contracts.internal.ImplementationKind.CALLABLE
        else:
            return contracts.internal.ImplementationKind.STATIC

    @staticmethod
    def _derive_implementation_execution_type(
        implementation: contracts.ImplementationType
    ):
        if inspect.iscoroutinefunction(implementation):
            return contracts.internal.ImplementationExecutionType.ASYNC
        elif inspect.isfunction(implementation):
            return contracts.internal.ImplementationExecutionType.SYNC
        elif hasattr(implementation, '__call__'):
            if inspect.iscoroutinefunction(implementation.__call__):
                return contracts.internal.ImplementationExecutionType.ASYNC
            else:
                return contracts.internal.ImplementationExecutionType.SYNC



import collections.abc
import typing
import dataclasses
import enum

from .domain import (
    ImplementationExecutionTrigger,
    CacheScope,
    ContractType,
    ImplementationType
)

InstanceType = typing.Any


class ImplementationKind(enum.Enum):
    CALLABLE = enum.auto()
    CLASS = enum.auto()
    STATIC = enum.auto()
    CALLBACK = enum.auto()


class ImplementationExecutionType(enum.Enum):
    SYNC = enum.auto()
    ASYNC = enum.auto()


@dataclasses.dataclass
class FunctionParameter:
    type: type
    default: typing.Any | None


@dataclasses.dataclass
class FunctionSignature:
    parameters: collections.abc.Mapping[str, FunctionParameter]
    result: type


@dataclasses.dataclass
class ContainerUnit:
    contract: ContractType
    implementation: ImplementationType
    implementation_kind: ImplementationKind
    implementation_execution_type: ImplementationExecutionType | None = None
    implementation_execution_trigger: ImplementationExecutionTrigger | None = None
    cache_scope: CacheScope | None = None
    init_signature: FunctionSignature | None = None
    execution_signature: FunctionSignature | None = None


ContainerUnitRegistry = collections.abc.MutableMapping[ContractType, ContainerUnit]
InstanceCacheType = collections.abc.MutableMapping[ContractType, InstanceType]



import typing
import enum
import collections.abc

ContractType = typing.Any
ImplementationType = typing.Any

CallbackContextType = collections.abc.Mapping[str, typing.Any]
CallbackType = collections.abc.Callable[[ContractType, CallbackContextType], typing.NoReturn]


class ImplementationExecutionTrigger(enum.Enum):
    SHUTDOWN = enum.auto()


class CacheScope(enum.Enum):
    CONTAINER = enum.auto()
    RESOLVER = enum.auto()


class ShutdownCallbackContext(typing.TypedDict):
    exc_type: type[Exception]
    exc_val: Exception
